#include <stdio.h>
#include <MClib.h>

#define MATRIX_SIZE 500

int a[MATRIX_SIZE][MATRIX_SIZE];
int b[MATRIX_SIZE][MATRIX_SIZE];
int c[MATRIX_SIZE][MATRIX_SIZE];

static unsigned long next = 1;
static unsigned long long int start_clock;

int myrand(void) {
    next = next * 1103515245 + 12345;
    return((unsigned)(next/65536) % 32768);
}

void init(int id_x, int id_y)
{
    if (id_x == 1 && id_y == 1) {
        int i, j;
        for (i = 0; i < MATRIX_SIZE; i++) {
            for (j= 0; j < MATRIX_SIZE; j++) {
                a[i][j] = myrand();
                b[i][j] = myrand();
            }
        }
        MC_clock(&start_clock);
    }
}

void fin(int id_x, int id_y)
{
    if (id_x == 1 && id_y == 1) {
        unsigned long long int clock;
        MC_clock(&clock);
        printf("cycle: %llu\n", clock - start_clock);
    }
}

void matrixprint(int a[MATRIX_SIZE][MATRIX_SIZE])
{
    int i, j;
    for (i = 0; i < MATRIX_SIZE; i++) {
        for (j= 0; j < MATRIX_SIZE; j++) {
            printf("%05d ", a[i][j]);
        }
        printf("\n");
    }
}

void debugprint(int id_x, int id_y)
{
    if (id_x == 1 && id_y == 1) {
        printf("a\n");
        matrixprint(a);
        printf("b\n");
        matrixprint(b);
        printf("c\n");
        matrixprint(c);
    }
}

volatile int start_flag;
volatile int end_flags[16][16];

int main()
{
    int id_x, id_y, rank_x, rank_y;
    MC_init(&id_x, &id_y, &rank_x, &rank_y);

    init(id_x, id_y);
    
    int smw = MATRIX_SIZE / rank_x; /* sub matrix width */
    int smh = MATRIX_SIZE / rank_y; /* sub matrix height */

    if (id_x == 1 && id_y == 1) {
        int x, y;
        for (y = 0; y < rank_y; y++) {
            for (x = 0; x < rank_x; x++) {
                int mcid;
                setidxy(&mcid, x + 1, y + 1);
                int yy;
                for (yy = y * smh; yy < y * smh + smh; yy++) {
                    MC_dma_put(mcid, &a[yy][x * smw], &a[yy][x * smw], sizeof(int) * smw, 4, 4);
                    MC_dma_put(mcid, &b[yy][x * smw], &b[yy][x * smw], sizeof(int) * smw, 4, 4);
                }
                MC_dma_put_4b(mcid, (void*)&start_flag, 1);
            }
        }
    }

    while (start_flag == 0);

    int x, y;
    int begin_y = (id_y - 1) * smh;
    int end_y   = begin_y + smh;
    int begin_x = (id_x - 1) * smw;
    int end_x   = begin_x + smw;
    int mcid;
    setidxy(&mcid, 1, 1);
    for (y = begin_y; y < end_y; y++) {
        for (x = begin_x; x < end_x; x++) {
            c[y][x] = a[y][x] + b[y][x];
        }
        MC_dma_put(mcid, &c[y][begin_x], &c[y][begin_x], sizeof(int) * smw, 4, 4);
    }
    MC_dma_put_4b(mcid, (void*)&end_flags[id_y-1][id_x-1], 1);

    if (id_x == 1 && id_y == 1) {
        for (y = 0; y < rank_y; y++) {
            for (x = 0; x < rank_x; x++) {
                while (end_flags[y][x] == 0);
            }
        }
    }

    fin(id_x, id_y);
    //debugprint(id_x, id_y);

    MC_finalize();
    return 0;
}
