/**********************************************************************/
/* Benchmark for Many-core processors            Arch Lab. TOKYO TECH */
/**********************************************************************/

#ifdef SC
#include "sc/MClib.h"
#else

#ifndef _MCLIB_H_
#define _MCLIB_H_

#include <stdio.h>
#include <stdarg.h>

#define GREAT_BARRIER
#define GREAT_REDUCE

/************************************************************************/
/* MieruPC Memory Map : 0x100 - 0x1ff                    Ver.2009-06-19 */
/************************************************************************/
#define MP_MMAP_BASE                   0x00000100

#define MP_MMAP_KEYCODE                (MP_MMAP_BASE + 0x00)
#define MP_MMAP_LCD                    (MP_MMAP_BASE + 0x04)
#define MP_MMAP_BUZZER                 (MP_MMAP_BASE + 0x08)
#define MP_MMAP_TIMER                  (MP_MMAP_BASE + 0x0c)
#define MP_MMAP_GPIO_0                 (MP_MMAP_BASE + 0xf0)
#define MP_MMAP_GPIO_4                 (MP_MMAP_BASE + 0xf4)
#define MP_MMAP_GPIO_8                 (MP_MMAP_BASE + 0xf8)
#define MP_MMAP_GPIO_C                 (MP_MMAP_BASE + 0xfc)

/************************************************************************/
/* ScalableCore Memory Map : 0x200 - 0x3ff               Ver.2009-06-19 */
/************************************************************************/
#define SC_MMAP_BASE                   0x00000200

#define SC_MMAP_NODE_ID                (SC_MMAP_BASE + 0x00)
#define SC_MMAP_NOC_SIZE               (SC_MMAP_BASE + 0x04)

#define SC_MMAP_PC                     (SC_MMAP_BASE + 0x08)
#define SC_MMAP_TIMER                  (SC_MMAP_BASE + 0x10) // clock
#define SC_MMAP_TIMER_D                (SC_MMAP_BASE + 0x14) // 

#define SC_MMAP_INST_LOOP              (SC_MMAP_BASE + 0x18) // ??? b
#define SC_MMAP_INST_LOOP__            (SC_MMAP_BASE + 0x1c) // ??? nop

#define SC_MMAP_DMA_PUT_REMOTE_ID      (SC_MMAP_BASE + 0x20)
#define SC_MMAP_DMA_PUT_REMOTE_ADDR    (SC_MMAP_BASE + 0x24)
#define SC_MMAP_DMA_PUT_REMOTE_STRIDE  (SC_MMAP_BASE + 0x28)
#define SC_MMAP_DMA_PUT_LOCAL_ADDR     (SC_MMAP_BASE + 0x2c)
#define SC_MMAP_DMA_PUT_LOCAL_STRIDE   (SC_MMAP_BASE + 0x30)
#define SC_MMAP_DMA_PUT_SIZE           (SC_MMAP_BASE + 0x34)
#define SC_MMAP_DMA_PUT_CMD            (SC_MMAP_BASE + 0x3c)

#define SC_MMAP_DMA_GET_REMOTE_ID      (SC_MMAP_BASE + 0x40)
#define SC_MMAP_DMA_GET_REMOTE_ADDR_L  (SC_MMAP_BASE + 0x44)
#define SC_MMAP_DMA_GET_REMOTE_ADDR_H  (SC_MMAP_BASE + 0x48)
#define SC_MMAP_DMA_GET_LOCAL_ADDR     (SC_MMAP_BASE + 0x4c)
#define SC_MMAP_DMA_GET_SIZE           (SC_MMAP_BASE + 0x50)
#define SC_MMAP_DMA_GET_REMOTE_STRIDE  (SC_MMAP_BASE + 0x54)
#define SC_MMAP_DMA_GET_LOCAL_STRIDE   (SC_MMAP_BASE + 0x58)
#define SC_MMAP_DMA_GET_CMD            (SC_MMAP_BASE + 0x5c)

#define SC_MMAP_02_N_TX                (SC_MMAP_BASE + 0x60)
#define SC_MMAP_02_E_TX                (SC_MMAP_BASE + 0x64)
#define SC_MMAP_02_W_TX                (SC_MMAP_BASE + 0x68)
#define SC_MMAP_02_S_TX                (SC_MMAP_BASE + 0x6c)
#define SC_MMAP_02_N_RX                (SC_MMAP_BASE + 0x70)
#define SC_MMAP_02_E_RX                (SC_MMAP_BASE + 0x74)
#define SC_MMAP_02_W_RX                (SC_MMAP_BASE + 0x78)
#define SC_MMAP_02_S_RX                (SC_MMAP_BASE + 0x7c)

#ifdef GREAT_BARRIER
# define SC_MMAP_GREAT_BARRIER         (SC_MMAP_BASE + 0x60)
#endif

#ifdef GREAT_REDUCE
#define SC_MMAP_GREAT_REDUCE_SADDR      (SC_MMAP_BASE + 0x64)
#define SC_MMAP_GREAT_REDUCE_RADDR      (SC_MMAP_BASE + 0x68)
#define SC_MMAP_GREAT_REDUCE_TYPE      (SC_MMAP_BASE + 0x6c)
#define SC_MMAP_GREAT_REDUCE_OP        (SC_MMAP_BASE + 0x70)
#define SC_MMAP_GREAT_REDUCE_SIZE      (SC_MMAP_BASE + 0x74)
#endif

#define SC_MMAP_DMA_FLAG               (SC_MMAP_BASE + 0x80)

/********************************************************************/
/* mem addr */
#define MEM_BASE                  0x0000200

#define GMEM_DMA_GET_REMOTE_ID     (MEM_BASE + 0x40)
#define GMEM_DMA_GET_REMOTE_ADDR_L (MEM_BASE + 0x44)
#define GMEM_DMA_GET_REMOTE_ADDR_H (MEM_BASE + 0x48)
#define GMEM_DMA_GET_LOCAL_ADDR_L  (MEM_BASE + 0x4c)
#define GMEM_DMA_GET_LOCAL_ADDR_H  (MEM_BASE + 0x50)
#define GMEM_DMA_GET_SIZE          (MEM_BASE + 0x54)
#define GMEM_DMA_GET_REMOTE_STRIDE (MEM_BASE + 0x58)
#define GMEM_DMA_GET_CMD           (MEM_BASE + 0x5c)

/* reserved address */
#define MEM_BARRIER_WORKER_FLAG   (MEM_BASE + 0xe0)    // 4byte
#define MEM_LOCK_WORKER_FLAG      (MEM_BASE + 0xe8)    // 4byte
#define MEM_LOCK_SERVER_FLAG      (MEM_BASE + 0xec)    // 4byte

#define MEM_BARRIER_SERVER_BASE   (MEM_BASE + 0x10000) // 4*cores
#define MEM_LOCK_SERVER_BASE      (MEM_BASE + 0x20000) // 4*cores

/********************************************************************/
#define CORE_LOCATE_BARRIER_X 1  // x location of the barrier server
#define CORE_LOCATE_LOCK_X    2  // x location of the lock server
#define CORE_LOADER           4

typedef unsigned int uint032_t;
typedef unsigned long long int uint064_t;

enum {
    DMA_NA = 0,
    DMA_PUT = 1,
    DMA_PUT64 = 2,
};

enum {
    DMAFLAG_NA,
    DMAFLAG_ACK,
    DMAFLAG_NACK,
};

typedef enum {
    MC_CHAR,
    MC_INT,
    MC_FLOAT,
    MC_DOUBLE,

    MC_CHARACTER = MC_CHAR,
    MC_INTEGER = MC_INT,
    MC_REAL = MC_FLOAT,
    MC_DOUBLE_PRECISION = MC_DOUBLE,
} MC_Datatype;

typedef enum {
    MC_MAX,
    MC_MIN,
    MC_SUM,
    MC_PROD,
    MC_LAND,
    MC_BAND,
    MC_LOR,
    MC_BOR,
    MC_LXOR,
    MC_BXOR,
    MC_MAXLOC,
    MC_MINLOC,
} MC_Op;

inline int MC_barrier();
int MC_server_barrier();
inline int MC_lock(int lock);
inline int MC_unlock(int lock);
int MC_server_lock();
int MC_server_loader();
inline void setidx(int*, int);
inline void setidy(int*, int);
inline void setidxy(int*, int, int);
inline int getidx(int);
inline int getidy(int);
inline int MC_get_id(int *id);
inline int MC_get_rank(int *rank);
inline void MC_initialize();
int MC_init(int *id_x, int *id_y, int *rank_x, int *rank_y);
void MC_finalize();
// There is no type 'bool' on C
// bool MC_dma_put_enable();
int MC_dma_put_enable();
void MC_dma_put(int dst_id, void *dst, void *src, size_t size,
                int dst_stride, int src_stride);
void MC_dma_put_nowait(int dst_id, void *dst, void *src, size_t size,
                       int dst_stride, int src_stride);
void MC_dma_puts(void *remote, void *local, size_t size,
                 int remote_stride, int local_stride);
void MC_dma_put64(int dst_id, uint064_t dst, void *src, size_t size,
                  int src_stride);
void MC_dma_put_4b(int dst_id, void *addr, int data);
void MC_dma_put_4b_nowait(int dst_id, void *addr, int data);
void MC_dma_get(int src_id, volatile void *dst, volatile void *src,
                size_t size,
    int dst_stride, int src_stride);
void MC_dma_get64(int remote_id, volatile uint064_t remote,
                  volatile void *local, size_t size, int local_stride);
int MC_sleep(long long unsigned int n);
int MC_clock(long long unsigned int *);
int MC_infoloop();
int MC_fast_barrier();
int MC_infloop();
void new_barrier();
void MC_gettime(unsigned int *t);

#ifdef GREAT_BARRIER
void MC_great_barrier();
#endif

#ifdef GREAT_REDUCE
void MC_great_reduce(void *send, void *recv,
                     MC_Datatype type, MC_Op op, size_t size);
#endif

#endif

#endif
