/**********************************************************************
 * SimMipsF: Simple Computer Simulator of MIPS   Arch Lab. TOKYO TECH *
 **********************************************************************/
enum {
    NFREG = 32,
    FIR_VALUE = 0x01538001,
};

/**********************************************************************/
typedef union {
    uint032_t w[NFREG];
    uint064_t l[NFREG / 2];
    float     s[NFREG];
    double    d[NFREG / 2];
} fpureg_t;

/**********************************************************************/
class MipsFpu {
 private:
    Board *board;
    fpureg_t r;
    uint032_t fcsr;
    char mnemonic[MNEMONIC_BUF_SIZE];
    char *instname;

 public:
    MipsFpu(Board *);
    void decode(MipsInst *);
    char *getinstname(MipsInst *);
    char *getmnemonic(MipsInst *);
    int execute(MipsInst *, MipsCp0 *, uint032_t *,
                uint032_t *, uint032_t *, uint032_t *,
                uint032_t *, uint064_t *, uint064_t *, int *);
    void writeback(MipsInst *, uint032_t, uint064_t);
    void print();
};


/* FPU Register Definition ********************************************/
/**********************************************************************/
enum {
    FPU_FIR_ = 0,
    FPU_FCCR = 25,
    FPU_FEXR = 26,
    FPU_FENR = 28,
    FPU_FCSR = 31,
};

/* FPU Instruction Definition *****************************************/
/**********************************************************************/
enum {
    MOVF_____,
    MOVT_____,
    MFC1_____,
    CFC1_____,
    MTC1_____,
    CTC1_____,
    BC1F_____,
    BC1T_____,
    BC1FL____,
    BC1TL____,
    ADD_S____,
    ADD_D____,
    SUB_S____,
    SUB_D____,
    MUL_S____,
    MUL_D____,
    DIV_S____,
    DIV_D____,
    SQRT_S___,
    SQRT_D___,
    ABS_S____,
    ABS_D____,
    MOV_S____,
    MOV_D____,
    NEG_S____,
    NEG_D____,
    ROUND_W_S,
    ROUND_W_D,
    TRUNC_W_S,
    TRUNC_W_D,
    CEIL_W_S_,
    CEIL_W_D_,
    FLOOR_W_S,
    FLOOR_W_D,
    MOVF_S___,
    MOVF_D___,
    MOVT_S___,
    MOVT_D___,
    MOVZ_S___,
    MOVZ_D___,
    MOVN_S___,
    MOVN_D___,
    RECIP_S__,
    RECIP_D__,
    RSQRT_S__,
    RSQRT_D__,
    CVT_S_D__,
    CVT_S_W__,
    CVT_D_S__,
    CVT_D_W__,
    CVT_W_S__,
    CVT_W_D__,
    C_F_S____,
    C_F_D____,
    C_UN_S___,
    C_UN_D___,
    C_EQ_S___,
    C_EQ_D___,
    C_UEQ_S__,
    C_UEQ_D__,
    C_OLT_S__,
    C_OLT_D__,
    C_ULT_S__,
    C_ULT_D__,
    C_OLE_S__,
    C_OLE_D__,
    C_ULE_S__,
    C_ULE_D__,
    C_SF_S___,
    C_SF_D___,
    C_SEQ_S__,
    C_SEQ_D__,
    C_NGL_S__,
    C_NGL_D__,
    C_LT_S___,
    C_LT_D___,
    C_NGE_S__,
    C_NGE_D__,
    C_LE_S___,
    C_LE_D___,
    C_NGT_S__,
    C_NGT_D__,
    LWXC1____,
    LDXC1____,
    SWXC1____,
    SDXC1____,
    MADD_S___,
    MADD_D___,
    MSUB_S___,
    MSUB_D___,
    NMADD_S__,
    NMADD_D__,
    NMSUB_S__,
    NMSUB_D__,
    L_S______,
    L_D______,
    S_S______,
    S_D______,
    UNDEF_FPU,
};

/**********************************************************************/
